/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- ui ------------------------- 20190913 - 332 / 283 vars


*** modular and extensible (graphical) ui framework - base module.


The ```ui``` framework or base takes care of most of the plumbing and data
flow between touch screen (or physical buttons) and display as input and
output components. ```ui``` is currently implemented as singleton. At runtime
it does not only hold on to its own code and data but also the code and data
of the on demand added and created ui elements.

Currently, plain buttons, check box, radio button, slider, input field - and
as last of course - a keyboard are available. Extensions are easily to create
and integrate based on the architecture and the available documentation.

All ui elements handled by ui framework have roughly 12 (13) common properties in
same and similar structure(s):

 - flags - hidden/displayed, touch active, read/display-only, in focus,...
 - clazz - btn-Button, chk-Checkbox, rad-Radiobutton, sli-Slider, inp-Inputfield,...
 - id of ui element - enabling sharing callbacks and other nifty things
 - x and y position - left-top of ui elements boundng box
 - width and height
 - border and fill colors
 - a value (object) enabling sharing callbacks and other nifty things
 - a preferred callback - on untouch
 - some label object(s) - plain, static text but also as renders (functions)
 - (optional, NON-preferred callback (on touch down) - only where applicable)

Only the 1st 9 of the 13 properties are really mandatory. Same values in same
structures have lead to an implementation of the ui elements as lightweight,
lean 'array objects' in order to save memory (variables).

The callback arguments include all pertinent information to allow a straight
forward use in the application and keep the code terse (to save memory / vars):

 - id of ui element
 - value (oject) of ui element
 - ui singleton (module) for accessing all statuses / ui elements / ui aspects
 - ui element subject to touch event (for optional query/change by application)
 - touch event object (touch/untouch and position on display / ui element)


*** Basic Application of ui base and - as example - use of uiBtn (Button) elements

--- get ui base and ui elements as needed, such as uiBtn, uiSli(der), ui...
Get ui base and ui elements and build ui from ui definitions in level 0 to
save vars, where it is executed on Espruino immediately on upload and this
ui definition source code is not stored. save() will save the built ui and
start it in the onInit() function on power up.

```
var ui = require("ui")       // getting ui base and control code (default 8 colors)
      .add(require("uiBtn")) // add/mixin btn ui elt support (and remove from cache)
      ;
```

--- define UI and add elements on code upload - but do not display it yet
When UI is built in level 0, build happens at upload to the board and source
code of UI definition is not consuming any runtime memory in order to save
variables. Example ui below defines ui with two (2) buttons: ```b01``` and
```b02```, which on release (untouch) will call the example callback ```cb``.
For details about this example ui see documentation about the uiBtn (button)
element.

```
//   0  1     2       3  4     5    6    7   8  9        10  11 
// flgs clazz id      x  y     w    h   bc  fc  valObj   cb,  l (label array obj)
//      btn                 ->x2 ->y2                          fv tc    x  y   text
ui.c(3,"btn","b01" ,  5, 40,  65,  35,  4,   4, "B_1",   cb,  [15, 7,  13, 9, "RED"  ]);
ui.c(3,"btn","b02" , 70, 40,  65,  35,  5,   6, {v:1},   cb,  [15, 0,   9, 9, "Y-w-B"]);
// bc/fc/tc = border/fill/text colors; cb=callback(id, v, ui, e, t). id is button id,
// such as "b01"; v, ui, e, and t provide button, ui, and touch event runtime data.
// colors 1-bit depth coded w/ 0b### ###=rgb (0=black, 7 is white, 4 is red, etc)
```

--- define application stuff - which is acting on ui events as callbacks. Example
callback ```cb`` - for simplicity reason - just logs in the console the buttons' id
and value (object).

```
function cb(id,v,_,e,t) { console.log(id + ": " + v); } // sample callback
```

--- do devices stuff - done on upload - once - defines the display and the touch
instance and module variable and set the module variable accordingly.

The ```ILI9341``` display controller as used in the examples is a 240 x 320 pixel,
max 16 bit color, TFT display controller. Other controller can be used as well with
adjusted with and height values and color definitions.

If the touch screen has a controller, use it's module - like ```XPT2646``` (goes
also for the ```ADS7843``` controller); otherwise use the module that can handle a
resistive touch screen directly (TouchRD: from Touchscreen, Resistive, plate
wires Directly connected to Espruino pins and controlled by Espruino).

```
var dsp,   dspMod   = require("ILI9341"); // display instance and module variable

var touch, touchMod = require("XPT2046"); // (or "ADS7843") touch screen module
  // or:
var touch, touchMod = require("TouchRD"); // touch screen w/o controller
```

--- define the onInit() function - it gets everything initialized, connected and
started. The code below shows module for ```XPT2046``` touch screen controller 
connection (can also be used for equal specified ```ADS7843``` controller) module
for touch screen connection.


```
function onInit() {
  SPI2.setup({sck:B13, miso:B14, mosi:B15, baud: 1000000});
  dsp = dspMod.connect(SPI2, B10, B1,  A4, function() {
    dsp.clear(); B2.set();   // clear display and switch backlight on
    ui.connect(dsp)          // connect ui to dsp (Espruino Graphics object)
      .w(0, 0, 240-1, 320-1) // wipe (clear/fill) screen w/ screen backgnd color
      .d()                   // display all ui elements
      .di = true;            // set to instant display
    SPI1.setup({sck:A5, miso:A6, mosi:A7, baud: 2000000}); // w/ touch controller...
    touch = touchMod.connect(SPI1, A3,  A2, function(x,y){ ui.evt(x,y);} // ...w/...
        , function(yr, xr, d) { // ...calirated raw to display x/y pos map function
            return [ Math.round(xr / -121.44          + 259.70685111989) 
                   , Math.round(yr /   88.90357142857 + -19.78130398103)
                   ];
        // or:
    touch = touchMod.connect(A6,A7,C2,C3,function(x,y){ ui.evt(x,y); }); // direct
    touchr.listen(true);   // not applicable for ADS7843 module
  });
} // /onInit()
```


--- start the code after upload
After uploading the code, enter ```onInit()``` in the Espruino IDE console to get
the code started. For convenience you can add the following line as last line in the
code, which gets the code automatically going after upload has successfully completed.
Note though to remove or comment this line on the last upload before saving the code
with ```save()```:

```
setTimeout(onInit,999); // for dev; remove on upload for save()
```


*** Description of (graphical) ui base module and touch controller interface

--- Functions / methods of ui base module:
ui provides also convenience functions/methods to be used in application to deal with
ui elements as well as plain graphics. Due to ui's extensibility, you can add easily
your own ui element or plain function extensions as desired (with .add(module, keep);).

- .iib(x,y,e) returns true when x and y (of touch) lay in bounding box of ui element e
- .foc(e) set focus to element e (and takes it away from element that had it so far)
- .blr(e) blur/unfocus element e (if .blr(), element currently in focus is blurred)
- .w(x,y,w,h,c) wipe rectangular area at x/y and w/h of display with color c
- .c(arguments) generic, single point error for creating ui elements. Uses first
     arguments[1] as class or type name and concatenates it with "C" for specific
     create entry. To create a custom ui element and integrate it (with
     .add(require("customUiElementModule")) is discussed in a separate publication.
- .d(e) redraw element e, and when just .d(), redraw the whole ui (all ui elements). In
     general, a ui is built on upload without displaying it to save variables and 
     displayed on demand or in onInit() to re-display on every re-power up or reset.
- .clr(c,i) set color according this.clrs{0](c,i) (customizable) function. Default is
     set for bit coded value for 1-bit (rgb) color depth and provides 8 colors:
     b&000=0=black, b&111=7=white, b&100=4=red, etc. this.clrs[0]() can be customized
     to support any color depth - coded, table looked up, or literal. For details see
     Colors section.
- .fnt(fv) set fontVector (size) ...currently w/ no frills, as opposed to .clr().
     More font options will be implemented at a later time. 
- .ld(x,y,l(abelInfoArray)) label draw function; label info array:
     l[0] : fontVector (size)
     l[1] : color (see Colors section
     l[2] : xOffset to x - left of bounding box of label (.drawString(...))
     l[3] : yOffset to y - top  of bounding box of label (.drawString(...))
     l[4] : label text (used in .drawString(...))
     Note: the reason for x and xOffset - and y and yOffset - is because .ld() is
     used to in conjunction with an ui element that has a bounding box that provides
     x and y - usually right-top corner - and and xOffset and yOffset are used to
     place the label relative to the ui elements supplied corner coordinates.
- .add(module,keep) extend ui module with ui element and other extensions to support
     modularity. After module code is merged into ui base code, module is removed from
     cache to save variables. For that purpose, the module has to know its name in mn
     property. For example, for a module onOff.js switching back light on/off on pin B2,
     exports = { mn:"onOff", on: function(on) { digitalWrite(B2, on || on===undefined); }
     defines the module code, ui.add(require("onOff")) adds it to be used as ui.on();.
     Without building a file, modules can also be put into cache directly on upload in 
     level 0 (and served with require() with module name provided as variable to escape 
     regular module detection and pre-upload from modules sources):
     Modules.addCached(mName,'{ mn:"onOff",on: function(on) { digitalWrite(B2,'
       +' on || on===undefined); }');, and added with short ui.add("onOff");

--- UI (touch) event object:
The touch event object ```t``` and other objects, such as the ```ui``` / ```_``` and
the ui element ```e``` themselves, are passed around for easy access in the ui base and
ui element implementation as well as in the application. The touch event object has this
structure:

```
t = // (touch) event
   { x: x // current or last touched x coordinate
   , y: y // current or last touched y coordinate
   , t: t // touching: (truey), not touching (anymore): falsey
   , d: d // touch down: (truey), on drag / untouch: falsey
   }
```

--- Colors and color definitions
ui base module includes a default color converter that supports 1-bit color depth
coded - single - RGB yielding 8 colors: 2^1^3 = 8 colors (3 Least Significant Bits
of an int). The color converter is a function and is first element - index [0] - of
an array which is set and stored as ```ui.clrs[]``` array property:

```
ui.clrs = [ function(c,i){ var v=(i)?c^7:c; return [v>>2&1,v>>1&1,v&1]; }.bind(exports) ]

```

The default color converter accepts for c a value from 0..7 for the 8 colors and
returns them as triplet of 0s and 1s (array of three (3) elements of values 0 and 1).
It accepts an optional second parameter to invert the color when truey. The inverted,
complementary color contrasts the non-inverted color and is defined by the color
specified by the 1-complement of color c. For example, for a the 1-bit color depth
defined color with value ```4```, ```[1r,0g,0b]``` or ```red```, the inverted,
contrasting complementary color value is ```3```, ```[0r,1g,1b]``` or ```aqua```.

The color converter is is invoked by ui's color setting method ```ui.clr(c,i)```
with same arguments. Any color converter delivers - has to deliver - the normalized
values 0..1 for each of the RGB colors as [r, g, b] triplet, because hat values
triplet is parameter for setting the color invoking the Espruino Graphics object's
set-color method ```.setColor(r, g, b)```. The display module knows how to pass the
color onto the display's controller to make the display show the proper color.

Using the enhanced color converter as below, colors can be specified in three ways:

- bit-coded as single positive value
- (negative) index into palette of [r,g,b] triplets
- literal [r,g,b] triplet

The n custom colors of the custom color palette are stored in the same array as the
color converter as 2nd to n-th + 1 element - index 1..n. The color converter has to
be built to accept 'negative' color values -1 .. -n to pick a color from the custom
palette: it takes the negative value as positive index into the ```ui.clrs[]```
array property tp pick a color from the custom color palette and return it as
value-normalized [0..1r,0..1g,0..1b] triplet. The way the custom colors of the
custom color palette are specified determines the implementation of the converter.

```
ui.clrs = // imply color depth and table w/ converter and define user colors / palette
  [function(c,i) { // color info to [r,g,b] converter at index 0 (ui.clrs[0](...))
       var v, s=isNaN(c); if (s || c<0) { // c<0 (looked up [R,G,B]) or c=[R,G,B]
       return (((s) ? c : this.clrs[-c]).map(function(v){ return ((i)?v^255:v)/256; }));
     } else { // c>=0 bit coded rgb color (default 1 bit color depth 2^1^3=8 colors)
       v = (i) ? c^7 : c; return [v>>2&1,v>>1&1,v&1]; } }.bind(ui)
  ,[216,216,216] // user color -1: light grey // keyboard bottom right special keys
  ,[192,192,192] // user color -2: medium light grey
  ,[128,128,128] // user color -3: medium gray
  ,[192,128,192] // user color -4: light purple
  ,[0  ,160, 64] // user color -5: kind of a 'darker' green
  ];
ui.bc = 0; // ui (screen) background color (override - with other value than default)
ui.tc = 7; // ui (screen) touch / focus color (override - with other value than default)
```

For [r,g,b] specification, Espruino Graphics modules accepts for each r,g,b base
color the normalized value 0..1. Under the hood and with help of the module for
the display, this normalized color specification is transformed - most of the
time - into a bit coded color again, but in a different, more elaborate /
complicated encoding, especially when different number of bits are used for each 
of the base r, g and b colors, for example, (display specific) 5, 6, 5 bits for
r, g, b. The module also handles the conversion the sequence of the r, g and b
colors when needed, for example, [r,g,b] to [r,b,g].

Normalized color specification value is mapped to actual color based on color
depth specified for or implied by the display (module) used in conjunction with
the Espruino's Graphics object. The ui base module already includes a default color
converter from bit coded rgb value to Espruino's normalized [r,g,b] values triplet.
Passing the normalized values triplets makes bit coded colors in the ui independent
from the display's version of bit coding. 

The n custom colors of the custom color palette are stored in the same array as the
color converter as 2nd to n-th + 1 element - index 1..n. The color converter has to
be built to accept 'negative' color values -1 .. -n to pick a color from the custom
palette: it takes the negative value as positive index into the ```ui.clrs[]```
array property tp pick a color from the custom color palette and return it as
value-normalized [0..1r,0..1g,0..1b] triplet. The way the custom colors of the
custom color palette are specified determines the implementation of the converter.

 */
exports = // ui base / 'DOM/ui e(lement) data & code holder, singleton, for mixins)
{ dsp: null // display (Espruino Graphics object)
, mn: "ui"  // reserved, used temporary
, bc: 0     // dsp background color >=0: r*g*b* bit coded; <0: looked up [...,[R,G,B],...]
, tc: 7     // touch / focus color; >=0: r*g*b* bit coded; <0: looked up [...,[R,G,B],...]
, di: false // display instantly on create/change (not defer until after save())
, es: []    // ui elements
, ef: null  // (ui) element (in) focus (by touch down)
, lf: null  // (ui element) last (in) focus (by touch down)
, af: null  // alternate (ui element in) focus (by drag over)
, it: false // (display) is touched
, td: false // is touch down (event)
, tt: 0     // touch down time
, lx: 0     // last x (touched)
, ly: 0     // last y (touched)
, dx: -1    // (last) x (touch) down
, dy: -1    // (last) y (touch) down
, te: {t:0} // (last) touch event
, clrs:     // default - bit coded color of 1 bit color-depth - 2^1^3(rgb) = 8 colors
    [function(c,i){ var v=(i)?c^7:c; return [v>>2&1,v>>1&1,v&1]; }.bind(exports)
    ]
, evt: function(x,y,_) { // common handling of touch / untouch / tap event
    _=_||this; var a=_.es, i=-1, m=a.length, e=_.ef, t={t:undefined!==x,d:0}; _.te=t;
    if (_.td=!_.it && t.t) { _.dx=x; _.dy=y; _.tt=getTime(); } // touch down
    if (_.it=t.t) { _.lx=t.x=x; _.ly=t.y=y; // touching
      if (e) { if (_.iib(e,x,y)) { return _[e[1]](_,e,t); } // xy event/move in focus elt
        _.blr(e); t.d=-1; _[e[1]](_,e,t); t={t:t.t,x:t.x,y:t.y,d:0}; } // just moved out of focused element *** added event call,-1,clone
      while (++i<m) { if (((e=a[i])[0]&2) && _.iib(e,x,y)) { // loop through active elts
        if (!(e[0]&4) && (_.td || (t.t && (e===_.lf)))) { _.foc(e); t.d=1; } // adj focus
        if (!(e[0]&12)) { e[0]|=8; t.d=1; }
        _[e[1]](_,e,t); return; } } if (_.td) { _.lf=null; } // handle hit / miss
    } else { if (e) { t.x=_.lx; t.y=_.ly; // untouch
      if (!_[e[1]](_,e,t,1)) { _.blr(e); } _.lf=null; _.dx=_.dy=-1; } } }
, iib: function(e,x,y) { // returns true if x/y is in bounding box of elt, else false
    return ((x>=e[3])&&(y>=e[4])&&(x<=e[5])&&(y<=e[6])); }
, foc: function(e,_) { // focus elt (unfocus elt in focus); visual: draw rectangle in...
    _=_||this; if (e!==_.ef && _.ef) { _.blr(_.ef); } // ...touch color 'around' elt
    _.clr(_.tc).dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]|=4; _.ef=_.lf=e; }
    return _; }
, blr: function(e,_) { // blur elt (all if e falsy); visual: draw rectangle in display...
    _=_||this; _.clr(_.bc); (e?[e]:_.es).forEach(function(e){ // ...background clr around
        if (e[0]&4) { _.dsp.drawRect(e[3],e[4],e[5],e[6]); e[0]&=11; } }); _.ef=null;
    return _; }
, rdc: function(e) { // register ui element, draw/display it conditionally, and return it
    this.es.push(e); if (((e[0]&=7)&1) && this.di) { this.d(e); }
    return e; }
, w: function(x,y,x2,y2,c) { // wipe rectangle (fill w/ c(olor) or display backgnd color)
    this.clr((c===undefined)?this.bc:c); this.dsp.fillRect(x,y,x2,y2); return this; }
, c: function() { // c(reate) ui element like DOM.createElement(), where arguments[1]...
    // ...defines ui element clazz and points to ui element specific constructor method
    var a=arguments, c=this[a[1]+"C"]; if (c) {
    return c.apply(this,a); } }  // undefined ui classes returns undefined as ui element
, d: function(e) { // display / redraw ui element e, redraw all if e falsy
    ((e)?[e]:this.es).forEach(function(e){ if (e[0]&1) this[e[1]+"D"].apply(this,e); },this);
    return this; }
, clr: function(c,i) { // set color according this.clrs[0](c,i) (customizable) function...
    this.dsp.setColor.apply(this.dsp,this.clrs[0](c,i)); // ...w/ color and i-nverse flag
    return this; }
, fnt: function(s) { // set displays's font vector (font size)
    this.dsp.setFontVector(s);
    return this; }
, ld: function(x,y,l) { // draw label (x, y, label[fv, tc, x, y, txt])
    this.fnt(l[0]).clr(l[1]).dsp.drawString(l[4],x+l[2],y+l[3]); }
, add: function(emon, k) { // add (mixin) non-overwriting properties of ui element by... 
    // eltModuleOrName for modularization and extensibility. Remove module from cache...
    var p, // ...when not k(eep) by element module name - if defined - to save space
    m = ("string" === typeof emon) ? require(emon) : emon; for (p in m) { if ((p!=="mn")
    &&!this[p]) { this[p] = m[p]; } } // merge / mixin properties 'new' (vars, functions)
    if (!k) { Modules.removeCached(m.mn); } // delete module from cache when not k(eep)
    return this; }
, connect: function(dsp) { // connects w/ initialized Espruino Grapics obj (after save()) 
    this.dsp=dsp;
    return this; }
};
